import uuid
from contextlib import asynccontextmanager
from typing import Any, Dict, Optional

from nonebot import get_driver
from nonebot.log import logger
from nonebot.utils import escape_tag
from playwright.async_api import Page, Route, async_playwright
from typing_extensions import Self

driver = get_driver()
try:
    import ujson as json
except ModuleNotFoundError:
    import json


SESSION_TOKEN_KEY = "__Secure-next-auth.session-token"


class Chatbot:
    def __init__(
        self,
        *,
        token: str = "",
        account: str = "",
        password: str = "",
        api: str = "https://chat.openai.com/",
        proxies: Optional[str] = None,
        timeout: int = 10,
    ) -> None:
        self.session_token = token
        self.account = account
        self.password = password
        self.api_url = api
        self.proxies = proxies
        self.timeout = timeout
        self.content = None
        self.parent_id = None
        self.conversation_id = None
        self.browser = None
        self.playwright = async_playwright()
        if self.session_token:
            self.auto_auth = False
        elif self.account and self.password:
            self.auto_auth = True
        else:
            raise ValueError("至少需要配置 session_token 或者 account 和 password")

    async def playwright_start(self):
        """启动浏览器，在插件开始运行时调用"""
        playwright = await self.playwright.start()
        try:
            self.browser = await playwright.firefox.launch(
                headless=True,
                proxy={"server": self.proxies} if self.proxies else None,  # your proxy
            )
        except Exception as e:
            logger.opt(exception=e).error("playwright未安装，请先在shell中运行playwright install")
            return
        ua = f"Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:72.0) Gecko/20100101 Firefox/{self.browser.version}"
        self.content = await self.browser.new_context(user_agent=ua)
        await self.set_cookie(self.session_token)

    async def set_cookie(self, session_token: str):
        """设置session_token"""
        self.session_token = session_token
        await self.content.add_cookies(
            [
                {
                    "name": SESSION_TOKEN_KEY,
                    "value": session_token,
                    "domain": "chat.openai.com",
                    "path": "/",
                }
            ]
        )

    @driver.on_shutdown
    async def playwright_close(self):
        """关闭浏览器"""
        await self.content.close()
        await self.browser.close()
        await self.playwright.__aexit__()

    def __call__(
        self, conversation_id: Optional[str] = None, parent_id: Optional[str] = None
    ) -> Self:
        self.conversation_id = conversation_id[-1] if conversation_id else None
        self.parent_id = parent_id[-1] if parent_id else self.id
        return self

    @property
    def id(self) -> str:
        return str(uuid.uuid4())

    def get_payload(self, prompt: str) -> Dict[str, Any]:
        return {
            "action": "next",
            "messages": [
                {
                    "id": self.id,
                    "role": "user",
                    "content": {"content_type": "text", "parts": [prompt]},
                }
            ],
            "conversation_id": self.conversation_id,
            "parent_message_id": self.parent_id,
            "model": "text-davinci-002-render",
        }

    @asynccontextmanager
    async def get_page(self):
        """打开网页，这是一个异步上下文管理器，使用async with调用"""
        page = await self.content.new_page()
        js = "Object.defineProperties(navigator, {webdriver:{get:()=>undefined}});"
        await page.add_init_script(js)
        await page.goto("https://chat.openai.com/chat")
        yield page
        await page.close()

    async def get_chat_response(self, prompt: str) -> str:
        async with self.get_page() as page:
            await page.wait_for_load_state("domcontentloaded")
            if not await page.locator("text=OpenAI Discord").is_visible():
                await self.get_cf_cookies(page)
            logger.debug("正在发送请求")

            async def change_json(route: Route):
                await route.continue_(
                    post_data=json.dumps(self.get_payload(prompt)),
                )

            await self.content.route(
                "https://chat.openai.com/backend-api/conversation", change_json
            )
            await page.wait_for_load_state("domcontentloaded")
            await page.wait_for_load_state("networkidle")
            session_expired = page.locator("button", has_text="Log in")
            if await session_expired.is_visible():
                logger.debug("检测到session过期")
                return "token失效，请重新设置token"
            next_botton = page.locator(
                ".btn.flex.justify-center.gap-2.btn-neutral.ml-auto"
            )
            if await next_botton.is_visible():
                logger.debug("检测到初次打开弹窗")
                await next_botton.click()
                await next_botton.click()
                await page.click(".btn.flex.justify-center.gap-2.btn-primary.ml-auto")
            async with page.expect_response(
                "https://chat.openai.com/backend-api/conversation",
                timeout=self.timeout * 1000,
            ) as response_info:
                textarea = page.locator("textarea")
                botton = page.locator('button[class="absolute p-1 rounded-md text-gray-500 bottom-1.5 right-1 md:bottom-2.5 md:right-2 hover:bg-gray-100 dark:hover:text-gray-400 dark:hover:bg-gray-900 disabled:hover:bg-transparent dark:disabled:hover:bg-transparent"]')
                logger.debug("正在等待回复")
                for _ in range(3):
                    await textarea.fill(prompt)
                    if await botton.is_enabled():
                        await botton.click()
                        break
                    await page.wait_for_timeout(500)
            response = await response_info.value
            if response.status == 429:
                return "请求过多，请放慢速度"
            if response.status == 403:
                await self.get_cf_cookies(page)
                return await self.get_chat_response(prompt)
            if response.status != 200:
                logger.opt(colors=True).error(
                    f"非预期的响应内容: <r>HTTP{response.status}</r> {escape_tag(response.text)}"
                )
                return f"ChatGPT 服务器返回了非预期的内容: HTTP{response.status}\n{response.text}"
            lines = await response.text()
            lines = lines.splitlines()
            data = lines[-4][6:]
            response = json.loads(data)
            self.parent_id = response["message"]["id"]
            self.conversation_id = response["conversation_id"]
            logger.debug("发送请求结束")
        return response["message"]["content"]["parts"][0]

    async def refresh_session(self) -> None:
        logger.debug("正在刷新session")
        if self.auto_auth:
            await self.login()
        else:
            async with self.get_page() as page:
                if not await page.locator("text=OpenAI Discord").is_visible():
                    await self.get_cf_cookies(page)
                await page.wait_for_load_state("domcontentloaded")
                session_expired = page.locator("text=Your session has expired")
                if await session_expired.count():
                    logger.opt(colors=True).error("刷新会话失败, session token 已过期, 请重新设置")
            cookies = await self.content.cookies()
            for i in cookies:
                if i["name"] == SESSION_TOKEN_KEY:
                    self.session_token = i["value"]
                    break
            logger.debug("刷新会话成功")

    async def login(self) -> None:
        from OpenAIAuth.OpenAIAuth import OpenAIAuth

        auth = OpenAIAuth(self.account, self.password, bool(self.proxies), self.proxies)  # type: ignore
        try:
            auth.begin()
        except Exception as e:
            if str(e) == "Captcha detected":
                logger.error("不支持验证码, 请使用 session token")
            raise e
        if not auth.access_token:
            logger.error("ChatGPT 登陆错误!")
        if auth.session_token:
            await self.set_cookie(auth.session_token)
        elif possible_tokens := auth.session.cookies.get(SESSION_TOKEN_KEY):
            if len(possible_tokens) > 1:
                await self.set_cookie(possible_tokens[0])
            else:
                try:
                    await self.set_cookie(possible_tokens)
                except Exception as e:
                    logger.opt(exception=e).error("ChatGPT 登陆错误!")
        else:
            logger.error("ChatGPT 登陆错误!")

    @staticmethod
    async def get_cf_cookies(page: Page) -> None:
        logger.debug("正在获取cf cookies")
        for _ in range(20):
            button = page.get_by_role("button", name="Verify you are human")
            if await button.count():
                await button.click()
            label = page.locator("label span")
            if await label.count():
                await label.click()
            await page.wait_for_timeout(1000)
            cf = page.locator("text=OpenAI Discord")
            if await cf.is_visible():
                break
        else:
            logger.error("cf cookies获取失败")
        logger.debug("cf cookies获取成功")



from pathlib import Path
from typing import List, Literal, Optional, Union

from nonebot import get_driver
from pydantic import BaseModel, Extra


class Config(BaseModel, extra=Extra.ignore):
    chatgpt_session_token: str = ""
    chatgpt_account: str = ""
    chatgpt_password: str = ""
    chatgpt_cd_time: int = 60
    chatgpt_proxies: Optional[str] = None
    chatgpt_refresh_interval: int = 30
    chatgpt_command: Union[str, List[str]] = ""
    chatgpt_to_me: bool = True
    chatgpt_timeout: int = 30
    chatgpt_api: str = "https://chat.openai.com/"
    chatgpt_image: bool = False
    chatgpt_image_width: int = 500
    chatgpt_priority: int = 999
    chatgpt_block: bool = True
    chatgpt_private: bool = True
    chatgpt_scope: Literal["private", "public"] = "private"
    chatgpt_data: Path = Path(__file__).parent
    chatgpt_max_rollback: int = 5
    chatgpt_detailed_error: bool = False


config = Config.parse_obj(get_driver().config)



from nonebot import on_command, require
from nonebot.adapters.onebot.v11 import (
    GroupMessageEvent,
    Message,
    MessageEvent,
    MessageSegment,
)
from nonebot.log import logger
from nonebot.params import CommandArg, _command_arg, _command_start
from nonebot.rule import to_me
from nonebot.typing import T_State

from playwright._impl._api_types import Error as PlaywrightAPIError

from .chatgpt import Chatbot
from .config import config
from .data import setting
from .utils import Session, cooldow_checker, create_matcher

require("nonebot_plugin_apscheduler")

from nonebot_plugin_apscheduler import scheduler

require("nonebot_plugin_htmlrender")

from nonebot_plugin_htmlrender import md_to_pic

#真寻帮助
__zx_plugin_name__ = "ChatGPT"
__plugin_usage__ = """
usage:
    官方ChatGPT插件
    指令：
        刷新会话/刷新对话 重置会话记录，开始新的对话
        导出会话/导出对话 导出当前会话记录
        导入会话/导入对话+会话ID+父消息ID(可选) 将会话记录导入，这会替换当前的会话
        保存会话/保存对话+会话名称 将当前会话保存
        查看会话/查看对话 查看已保存的所有会话
        切换会话/切换对话+会话名称 切换到指定的会话
        回滚会话/回滚对话 返回到之前的会话，输入数字可以返回多个会话，但不可以超过最大支持数量
""".strip()
__plugin_des__ = "ChatGPT"
__plugin_cmd__ = ["刷新回话","刷新对话","导出对话","导出对话","导入会话","保存会话","查看会话","查看对话","切换会话","回滚会话","回滚对话"]
__plugin_type__ = ("功能",)
__plugin_version__ = 1.0
__plugin_author__ = "INSide_734"
__plugin_settings__ = {
    "level": 5,
    "limit_superuser": False,
    "default_status": True,
    "cmd": ["刷新回话","刷新对话","导出对话","导出对话","导入会话","保存会话","查看会话","查看对话","切换会话","回滚会话","回滚对话"],
}

chat_bot = Chatbot(
    token=setting.token or config.chatgpt_session_token,
    account=config.chatgpt_account,
    password=config.chatgpt_password,
    api=config.chatgpt_api,
    proxies=config.chatgpt_proxies,
    timeout=config.chatgpt_timeout,
)

matcher = create_matcher(
    config.chatgpt_command,
    config.chatgpt_to_me,
    config.chatgpt_private,
    config.chatgpt_priority,
    config.chatgpt_block,
)

session = Session(config.chatgpt_scope)


def check_purview(event: MessageEvent) -> bool:
    return not (
        isinstance(event, GroupMessageEvent)
        and config.chatgpt_scope == "public"
        and event.sender.role == "member"
    )


@matcher.handle(parameterless=[cooldow_checker(config.chatgpt_cd_time)])
async def ai_chat(event: MessageEvent, state: T_State) -> None:
    if not chat_bot.content:
        await chat_bot.playwright_start()
    message = _command_arg(state) or event.get_message()
    text = message.extract_plain_text().strip()
    if start := _command_start(state):
        text = text[len(start):]
    try:
        msg = await chat_bot(**session[event]).get_chat_response(text)
        if (msg == "token失效，请重新设置token") and (
            chat_bot.session_token != config.chatgpt_session_token
        ):
            await chat_bot.set_cookie(config.chatgpt_session_token)
            msg = await chat_bot(**session[event]).get_chat_response(text)
    except PlaywrightAPIError as e:
        error = f"{type(e).__name__}: {e}"
        logger.opt(exception=e).error(f"ChatGPT request failed: {error}")
        if type(e).__name__ == "TimeoutError":
            await matcher.finish(
                "ChatGPT回复已超时。", at_sender=True
            )
        elif type(e).__name__ == "Error":
            msg = "ChatGPT 目前无法回复您的问题。"
            if config.chatgpt_detailed_error:
                msg += f"\n{error}"
            else:
                msg += "可能的原因是同时提问过多，问题过于复杂等。"
            await matcher.finish(
                msg, at_sender=True
            )
    if config.chatgpt_image:
        if msg.count("```") % 2 != 0:
            msg += "\n```"
        img = await md_to_pic(msg, width=config.chatgpt_image_width)
        msg = MessageSegment.image(img)
    await matcher.send(msg, at_sender=True)
    session[event] = chat_bot.conversation_id, chat_bot.parent_id


refresh = on_command("刷新对话", aliases={"刷新会话"}, block=True, rule=to_me(), priority=1)


@refresh.handle()
async def refresh_conversation(event: MessageEvent) -> None:
    if not check_purview(event):
        await import_.finish("当前为公共会话模式, 仅支持群管理操作")
    del session[event]
    await refresh.send("当前会话已刷新")


export = on_command("导出对话", aliases={"导出会话"}, block=True, rule=to_me(), priority=1)


@export.handle()
async def export_conversation(event: MessageEvent) -> None:
    if cvst := session[event]:
        await export.send(
            f"已成功导出会话:\n"
            f"会话ID: {cvst['conversation_id'][-1]}\n"
            f"父消息ID: {cvst['parent_id'][-1]}",
            at_sender=True,
        )
    else:
        await export.finish("你还没有任何会话记录", at_sender=True)


import_ = on_command(
    "导入对话", aliases={"导入会话", "加载对话", "加载会话"}, block=True, rule=to_me(), priority=1
)


@import_.handle()
async def import_conversation(event: MessageEvent, arg: Message = CommandArg()) -> None:
    if not check_purview(event):
        await import_.finish("当前为公共会话模式, 仅支持群管理操作")
    args = arg.extract_plain_text().strip().split()
    if not args:
        await import_.finish("至少需要提供会话ID", at_sender=True)
    if len(args) > 2:
        await import_.finish("提供的参数格式不正确", at_sender=True)
    session[event] = args.pop(0), args[0] if args else None
    await import_.send("已成功导入会话", at_sender=True)


save = on_command("保存对话", aliases={"保存会话"}, block=True, rule=to_me(), priority=1)


@save.handle()
async def save_conversation(event: MessageEvent, arg: Message = CommandArg()) -> None:
    if not check_purview(event):
        await save.finish("当前为公共会话模式, 仅支持群管理操作")
    if session[event]:
        name = arg.extract_plain_text().strip()
        session.save(name, event)
        await save.send(f"已将当前会话保存为: {name}", at_sender=True)
    else:
        await save.finish("你还没有任何会话记录", at_sender=True)


check = on_command("查看对话", aliases={"查看会话"}, block=True, rule=to_me(), priority=1)


@check.handle()
async def check_conversation(event: MessageEvent) -> None:
    name_list = "\n".join(list(session.find(event).keys()))
    await check.send(f"已保存的会话有:\n{name_list}", at_sender=True)


switch = on_command("切换对话", aliases={"切换会话"}, block=True, rule=to_me(), priority=1)


@switch.handle()
async def switch_conversation(event: MessageEvent, arg: Message = CommandArg()) -> None:
    if not check_purview(event):
        await switch.finish("当前为公共会话模式, 仅支持群管理操作")
    name = arg.extract_plain_text().strip()
    try:
        session[event] = session.find(event)[name]
        await switch.send(f"已切换到会话: {name}", at_sender=True)
    except KeyError:
        await switch.send(f"找不到会话: {name}", at_sender=True)


@scheduler.scheduled_job("interval", minutes=config.chatgpt_refresh_interval)
async def refresh_session() -> None:
    await chat_bot.refresh_session()
    setting.token = chat_bot.session_token
    setting.save()


rollback = on_command("回滚对话", aliases={"回滚会话"}, block=True, rule=to_me(), priority=1)


@rollback.handle()
async def rollback_conversation(event: MessageEvent, arg: Message = CommandArg()):
    num = arg.extract_plain_text().strip()
    if num.isdigit():
        num = int(num)
        if session[event]:
            count = session.count(event)
            if num > count:
                await rollback.finish(f"历史会话数不足，当前历史会话数为{count}", at_sender=True)
            else:
                for _ in range(num):
                    session.pop(event)
                await rollback.send(f"已成功回滚{num}条会话", at_sender=True)
        else:
            await save.finish("你还没有任何会话记录", at_sender=True)
    else:
        await rollback.finish(
            f"请输入有效的数字，最大回滚数为{config.chatgpt_max_rollback}", at_sender=True
        )



from pathlib import Path
from typing import Any, Dict

from pydantic import BaseModel, Field, root_validator

from .config import config

try:
    import ujson as json
except ModuleNotFoundError:
    import json


class Setting(BaseModel):
    session: Dict[str, Dict[str, Any]] = Field(default_factory=dict)
    token: str = ""

    __file_path: Path = config.chatgpt_data / "setting.json"

    @property
    def file_path(self) -> Path:
        return self.__class__.__file_path

    @root_validator(pre=True)
    def init(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if cls.__file_path.is_file():
            return json.loads(cls.__file_path.read_text("utf-8"))
        return values

    def save(self) -> None:
        self.file_path.write_text(self.json(), encoding="utf-8")


setting = Setting()



from collections import defaultdict, deque
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
)

from nonebot import on_command, on_message
from nonebot.adapters.onebot.v11 import GROUP, GroupMessageEvent, MessageEvent
from nonebot.matcher import Matcher
from nonebot.params import Depends
from nonebot.rule import to_me

from .config import config
from .data import setting


def cooldow_checker(cd_time: int) -> Any:
    cooldown = defaultdict(int)

    async def check_cooldown(
        matcher: Matcher, event: MessageEvent
    ) -> AsyncGenerator[None, None]:
        cooldown_time = cooldown[event.user_id] + cd_time
        if event.time < cooldown_time:
            await matcher.finish(
                f"ChatGPT 冷却中，剩余 {cooldown_time - event.time} 秒", at_sender=True
            )
        yield
        cooldown[event.user_id] = event.time

    return Depends(check_cooldown)


def create_matcher(
    command: Union[str, List[str]],
    only_to_me: bool = True,
    private: bool = True,
    priority: int = 999,
    block: bool = True,
) -> Type[Matcher]:
    params: Dict[str, Any] = {
        "priority": priority,
        "block": block,
    }

    if command:
        on_matcher = on_command
        command = [command] if isinstance(command, str) else command
        params["cmd"] = command.pop(0)
        params["aliases"] = set(command)
    else:
        on_matcher = on_message

    if only_to_me:
        params["rule"] = to_me()
    if not private:
        params["permission"] = GROUP

    return on_matcher(**params)


class Session(dict):
    def __init__(self, scope: Literal["private", "public"]) -> None:
        super().__init__()
        self.is_private = scope == "private"

    def __getitem__(self, event: MessageEvent) -> Dict[str, Any]:
        return super().__getitem__(self.id(event))

    def __setitem__(
        self,
        event: MessageEvent,
        value: Union[Tuple[Optional[str], Optional[str]], Dict[str, Any]],
    ) -> None:
        if isinstance(value, tuple):
            conversation_id, parent_id = value
        else:
            conversation_id = value["conversation_id"]
            parent_id = value["parent_id"]
        if self[event]:
            if isinstance(value, tuple):
                self[event]["conversation_id"].append(conversation_id)
                self[event]["parent_id"].append(parent_id)
        else:
            super().__setitem__(
                self.id(event),
                {
                    "conversation_id": deque(
                        [conversation_id], maxlen=config.chatgpt_max_rollback
                    ),
                    "parent_id": deque([parent_id], maxlen=config.chatgpt_max_rollback),
                },
            )

    def __delitem__(self, event: MessageEvent) -> None:
        sid = self.id(event)
        if sid in self:
            super().__delitem__(sid)

    def __missing__(self, _) -> Dict[str, Any]:
        return {}

    def id(self, event: MessageEvent) -> str:
        if self.is_private:
            return event.get_session_id()
        return str(
            event.group_id if isinstance(event, GroupMessageEvent) else event.user_id
        )

    def save(self, name: str, event: MessageEvent) -> None:
        sid = self.id(event)
        if setting.session.get(sid) is None:
            setting.session[sid] = {}
        setting.session[sid][name] = {
            "conversation_id": self[event]["conversation_id"][-1],
            "parent_id": self[event]["parent_id"][-1],
        }
        setting.save()

    def find(self, event: MessageEvent) -> Dict[str, Any]:
        sid = self.id(event)
        return setting.session[sid]

    def count(self, event: MessageEvent) -> int:
        return len(self[event]["conversation_id"])

    def pop(self, event: MessageEvent) -> Tuple[str, str]:
        conversation_id = self[event]["conversation_id"].pop()
        parent_id = self[event]["parent_id"].pop()
        return conversation_id, parent_id
